require "base64"
require 'optparse'
require_relative 'values_manager'
require_relative 'state_manager'

module Kerbi
  ##
  # Singleton class providing the user an interface with Kerbi
  class Engine
    ##
    # List of generators to produce hash-resources for YAMLification
    # @return [Array<Kerbi::Mixer>] array of generators
    attr_accessor :generators

    ##
    # Memoized values loaded by ValuesLoader
    # @return [SymbolHash] values tree to be passed to generators
    def values
      @_values ||= ValuesManager.load
    end

    def arg_values(name)
      indicies = ARGV.each_index.select { |i| ARGV[i]==name }
      indicies.map { |key_index| ARGV[key_index + 1] }
    end

    ##
    # Serializes all generators' outputs into a YAML string
    # @return [String] YAML string of all generator-produced resources
    def gen_yaml(values)
      self.gen(values).each_with_index.map do |h, i|
        raw = YAML.dump(h.deep_stringify_keys)
        raw.gsub("---\n", i.zero? ? '' : "---\n\n")
      end.join("\n")
    end

    def hash_to_printable_s(hash)
      raw = YAML.dump((hash || {}).deep_stringify_keys)
      raw.gsub("---\n", '')
    end

    ##
    # KTEA-adherent CLI arg routing
    # @return [void] Prints any output to stdout
    def cli_exec
      if ARGV[0] == 'template'
        puts self.gen_yaml(self.values)
      elsif ARGV[0..1] == %w[state patch]
        StateManager.new.patch
      elsif ARGV[0..1] == %w[state get]

      elsif ARGV[0..1] == %w[state template]
        state_values = StateManager.new.get_crt_vars
        merged_values = self.values.deep_merge(state_values)
        puts self.gen_yaml(merged_values)
      elsif ARGV[0..1] == %w[show values]
        puts hash_to_printable_s(self.values)
      elsif ARGV[0..1] == %w[show state]
        puts StateManager.new.get_crt_vars.to_json
      elsif ARGV[0..1] == %w[show preset]
        values = ValuesManager.safely_read_values_file(ARGV[2])
        puts hash_to_printable_s(values)
      else
        puts "Unrecognized command #{ARGV}"
        exit 1
      end
    end

    def gen(values)
      res_defs = self.generators.inject([]) do |whole, gen_class|
        generator = gen_class.new(values)
        whole + generator.run.flatten
      end

      if (filters = args_manager.get_res_filters).any?
        res_defs = res_defs.select do |res_def|
          res_def = res_def.deep_symbolize_keys
          kind = res_def[:kind]
          name = res_def.dig(:metadata, :name)
          positive_filters = filters.select do |filter|
            against_kind, against_name = filter.split(":")
            against_kind == kind && against_name == name
          end
          positive_filters.any?
        end
      end
      res_defs
    end
  end
end

##
# The singleton Kerbi::Engine, where the user can add
# generators invoke invoke the templating engine
# @return [Kerbi::Engine] singleton instance of Kerbi::App
def kerbi
  $kerbi ||= Kerbi::Engine.new
end

def tami_sleep
  while true
    sleep 10
  end
end
